from sssekai.unity.AssetBundle import load_assetbundle
import os, json


def main_mvdata(args):
    from UnityPy.enums import ClassIDType

    source = args.infile
    source = os.path.expanduser(source)
    source = os.path.abspath(source)
    os.chdir(source)
    mvdata_keys = sorted(os.listdir(source))
    mvdata_items = list()
    for key in mvdata_keys:
        try:
            with open(key, "rb") as f:
                env = load_assetbundle(f)
                for obj in env.objects:
                    if obj.type == ClassIDType.MonoBehaviour:
                        data = obj.read()
                        if data.m_Name == "data":
                            typetree = obj.read_typetree()
                            typetree = {
                                k: v
                                for k, v in typetree.items()
                                if not k.startswith("m_")
                            }
                            mvdata_items.append(typetree)
                            break
        except Exception as e:
            print(f"skipping {key}: {e}")
    outdir = os.path.dirname(args.outdir)
    if outdir:
        os.makedirs(outdir, exist_ok=True)
    with open(args.outdir, "w", encoding="utf-8") as f:
        json.dump(mvdata_items, f, indent=4, ensure_ascii=False)



from . import *


def test_live2d_motion():
    from sssekai.entrypoint.live2dextract import main_live2dextract

    result = main_live2dextract(
        NamedDict(
            {
                "infile": sample_file_path("live2d", "21miku_motion_base"),
                "outdir": TEMP_DIR,
            }
        )
    )


def test_live2d_model():
    from sssekai.entrypoint.live2dextract import main_live2dextract

    result = main_live2dextract(
        NamedDict(
            {
                "infile": sample_file_path("live2d", "21miku_night"),
                "outdir": TEMP_DIR,
            }
        )
    )


if __name__ == "__main__":
    test_live2d_motion()
    test_live2d_model()



from Crypto.Cipher import AES
from typing import Tuple


def PKCS7_pad(data: bytes, bs) -> bytes:
    size = bs - len(data) % bs
    return data + bytes([size] * size)


def PKCS7_unpad(data: bytes, bs) -> bytes:
    pad = data[-1]
    return data[:-pad]


def decrypt_aes_cbc(data: bytes, key, iv, unpad=PKCS7_unpad) -> bytes:
    cipher = AES.new(key, AES.MODE_CBC, iv)
    return unpad(cipher.decrypt(data), cipher.block_size)


def encrypt_aes_cbc(data: bytes, key, iv, pad=PKCS7_pad) -> bytes:
    cipher = AES.new(key, AES.MODE_CBC, iv)
    return cipher.encrypt(pad(data, cipher.block_size))


SEKAI_APIMANAGER_KEYSETS = {
    "en": (
        b"\xdf8B\x14\xb2\x9a:\xdf\xbf\x1b\xd9\xee[\x16\xf8\x84",
        b"~\x85l\x90y\x87\xf8\xae\xc6\xaf\xc0\xc5G8\xfc~",
    ),
    "jp": (b"g2fcC0ZczN9MTJ61", b"msx3IV0i9XE5uYZ1"),
    "tw": (b"g2fcC0ZczN9MTJ61", b"msx3IV0i9XE5uYZ1"),
    "kr": (b"g2fcC0ZczN9MTJ61", b"msx3IV0i9XE5uYZ1"),
    "cn": (b"g2fcC0ZczN9MTJ61", b"msx3IV0i9XE5uYZ1"),
}


def encrypt(data: bytes, keyset: Tuple[bytes, bytes]) -> bytes:
    return encrypt_aes_cbc(data, *keyset, PKCS7_pad)


def decrypt(data: bytes, keyset: Tuple[bytes, bytes]) -> bytes:
    return decrypt_aes_cbc(data, *keyset, PKCS7_unpad)



# Auto-generated by https://github.com/mos9527/UnityPyTypetreeCodegen
# Python definition for Sekai

from .. import *

@UTTCGen
class PlayerSettingConfig(MonoBehaviour):
	memo : str
	clientMajorVersion : str
	clientMinorVersion : str
	clientBuildVersion : str
	snapshot : str
	clientVersionSuffix : str
	clientDataMajorVersion : str
	clientDataMinorVersion : str
	clientDataBuildVersion : str
	clientDataRevision : str
	companyName : str
	productName : str
	bundleIdentifier : str
	bundleVersion : str
	assetHash : str
	clientAppHash : str
	adMobAppId : str



# fmt: off
# Auto-generated by https://github.com/mos9527/UnityPyTypetreeCodegen
from typing import List, Union, Optional, TypeVar
from UnityPy.classes import *
from UnityPy.classes.math import (ColorRGBA, Matrix3x4f, Matrix4x4f, Quaternionf, Vector2f, Vector3f, Vector4f, float3, float4,)
T = TypeVar("T")
def UTTCGen(clazz : T) -> T:
	"""dataclass-like decorator for typetree classess with nested type support
	
	limitations:
	- the behavior is similar to slotted dataclasses where shared attributes are inherited
	  but allows ommiting init of the parent if kwargs are not sufficient
	- generally supports nested types, however untested and could be slow	
    - and ofc, zero type checking and safeguards :/	
	"""
	RESERVED_KWS = {'object_reader'}
    # Allow these to be propogated to the props
	def __init__(self, **d):
		def enusre_reserved(obj):
			for k in RESERVED_KWS:
				if hasattr(obj, k) and k in d:
					setattr(obj, k, d[k])
			return obj		
		def reduce_init(clazz, **d):
			types : dict = clazz.__annotations__
			for k, sub in types.items():
				if type(sub) == str:
					sub = eval(sub) # attrs turns these into strings...why?
				reduce_arg = getattr(sub, "__args__", [None])[0]
				if isinstance(d[k], list):
					if hasattr(reduce_arg, "__annotations__"):
						setattr(self, k, [enusre_reserved(reduce_arg(**x)) for x in d[k]])
					else:
						setattr(self, k, [enusre_reserved(reduce_arg(x)) for x in d[k]])
				elif isinstance(d[k], dict) and hasattr(sub, "__annotations__"):
					setattr(self, k, enusre_reserved(sub(**d[k])))
				else:
					if isinstance(d[k], dict):
						setattr(self, k, enusre_reserved(sub(**d[k])))
					else:
						setattr(self, k, enusre_reserved(sub(d[k])))			
		for __base__ in clazz.__bases__:
			types : dict = __base__.__annotations__
			args = {k:d[k] for k in types if k in d}
			if len(args) == len(types):
				super(clazz, self).__init__(**args)
				reduce_init(__base__, **d)
				for k in args:
					if not k in RESERVED_KWS: del d[k]
		reduce_init(clazz, **d)
		enusre_reserved(self)
	def __repr__(self) -> str:
		return f"{clazz.__name__}({', '.join([f'{k}={getattr(self, k)!r}' for k in self.__annotations__])})"
	clazz.__init__ = __init__
	clazz.__repr__ = __repr__
	return clazz
    
# Helper functions
def UTTCGen_Reread(Mono: MonoBehaviour):
    script = Mono.m_Script.read()
    fullname = script.m_ClassName
    if script.m_Namespace:
        fullname = f"{script.m_Namespace}.{fullname}"
    definition = TYPETREE_DEFS.get(fullname, None)
    assert definition is not None, f"Typetree definition for {fullname} not found"
    raw_def = Mono.object_reader.read_typetree(definition)
    return raw_def, fullname

TYPETREE_DEFS = {
    "Sekai.PlayerSettingConfig": [
        {
            "m_Type": "MonoBehaviour",
            "m_Name": "Base",
            "m_Level": 0,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "PPtr<GameObject>",
            "m_Name": "m_GameObject",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "int",
            "m_Name": "m_FileID",
            "m_Level": 2,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "SInt64",
            "m_Name": "m_PathID",
            "m_Level": 2,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "UInt8",
            "m_Name": "m_Enabled",
            "m_Level": 1,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "PPtr<MonoScript>",
            "m_Name": "m_Script",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "int",
            "m_Name": "m_FileID",
            "m_Level": 2,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "SInt64",
            "m_Name": "m_PathID",
            "m_Level": 2,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "string",
            "m_Name": "m_Name",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "Array",
            "m_Name": "Array",
            "m_Level": 2,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "int",
            "m_Name": "size",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "char",
            "m_Name": "data",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "string",
            "m_Name": "memo",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "Array",
            "m_Name": "Array",
            "m_Level": 2,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "int",
            "m_Name": "size",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "char",
            "m_Name": "data",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "string",
            "m_Name": "clientMajorVersion",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "Array",
            "m_Name": "Array",
            "m_Level": 2,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "int",
            "m_Name": "size",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "char",
            "m_Name": "data",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "string",
            "m_Name": "clientMinorVersion",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "Array",
            "m_Name": "Array",
            "m_Level": 2,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "int",
            "m_Name": "size",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "char",
            "m_Name": "data",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "string",
            "m_Name": "clientBuildVersion",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "Array",
            "m_Name": "Array",
            "m_Level": 2,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "int",
            "m_Name": "size",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "char",
            "m_Name": "data",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "string",
            "m_Name": "snapshot",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "Array",
            "m_Name": "Array",
            "m_Level": 2,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "int",
            "m_Name": "size",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "char",
            "m_Name": "data",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "string",
            "m_Name": "clientVersionSuffix",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "Array",
            "m_Name": "Array",
            "m_Level": 2,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "int",
            "m_Name": "size",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "char",
            "m_Name": "data",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "string",
            "m_Name": "clientDataMajorVersion",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "Array",
            "m_Name": "Array",
            "m_Level": 2,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "int",
            "m_Name": "size",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "char",
            "m_Name": "data",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "string",
            "m_Name": "clientDataMinorVersion",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "Array",
            "m_Name": "Array",
            "m_Level": 2,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "int",
            "m_Name": "size",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "char",
            "m_Name": "data",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "string",
            "m_Name": "clientDataBuildVersion",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "Array",
            "m_Name": "Array",
            "m_Level": 2,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "int",
            "m_Name": "size",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "char",
            "m_Name": "data",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "string",
            "m_Name": "clientDataRevision",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "Array",
            "m_Name": "Array",
            "m_Level": 2,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "int",
            "m_Name": "size",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "char",
            "m_Name": "data",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "string",
            "m_Name": "companyName",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "Array",
            "m_Name": "Array",
            "m_Level": 2,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "int",
            "m_Name": "size",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "char",
            "m_Name": "data",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "string",
            "m_Name": "productName",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "Array",
            "m_Name": "Array",
            "m_Level": 2,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "int",
            "m_Name": "size",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "char",
            "m_Name": "data",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "string",
            "m_Name": "bundleIdentifier",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "Array",
            "m_Name": "Array",
            "m_Level": 2,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "int",
            "m_Name": "size",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "char",
            "m_Name": "data",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "string",
            "m_Name": "bundleVersion",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "Array",
            "m_Name": "Array",
            "m_Level": 2,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "int",
            "m_Name": "size",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "char",
            "m_Name": "data",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "string",
            "m_Name": "assetHash",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "Array",
            "m_Name": "Array",
            "m_Level": 2,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "int",
            "m_Name": "size",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "char",
            "m_Name": "data",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "string",
            "m_Name": "clientAppHash",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "Array",
            "m_Name": "Array",
            "m_Level": 2,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "int",
            "m_Name": "size",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "char",
            "m_Name": "data",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "string",
            "m_Name": "adMobAppId",
            "m_Level": 1,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "Array",
            "m_Name": "Array",
            "m_Level": 2,
            "m_MetaFlag": 16384
        },
        {
            "m_Type": "int",
            "m_Name": "size",
            "m_Level": 3,
            "m_MetaFlag": 0
        },
        {
            "m_Type": "char",
            "m_Name": "data",
            "m_Level": 3,
            "m_MetaFlag": 0
        }
    ]
}


import zipfile
import UnityPy
import logging
import re
import sys

from io import BytesIO

from UnityPy.enums import ClassIDType
from UnityPy.helpers.Tpk import get_typetree_node
from sssekai.unity.AssetBundle import load_assetbundle
from tqdm import tqdm

REGION_MAP = {
    # NOTE: Order is used to determine the region
    "com.hermes.mk": "cn",
    "com.hermes.mk.asia": "tw",
    "com.sega.ColorfulStage.en": "en",
    "com.sega.pjsekai": "jp",
    "com.pjsekai.kr": "kr",
}
logger = logging.getLogger("apphash")


def enum_candidates(zip_file, filter):
    return (
        (f, zip_file.open(f), zip_file) for f in zip_file.filelist if filter(f.filename)
    )


def enum_package(zip_file):
    yield zip_file
    for f in zip_file.filelist:
        if f.filename.lower().endswith(".apk"):
            yield zipfile.ZipFile(zip_file.open(f))


def dump_axml_stringpool(f: BytesIO):
    read_int = lambda nbytes: int.from_bytes(f.read(nbytes), "little")
    f.seek(8)
    hdr_type, hdr_size, size = read_int(2), read_int(2), read_int(4)
    n_strings = read_int(4)
    unk = read_int(4)
    flags = read_int(4)
    is_utf8 = flags & (1 << 8)
    string_offset = read_int(4) + 8
    unk = read_int(4)
    offsets = [read_int(4) for _ in range(n_strings)]
    for offset in offsets:
        f.seek(string_offset + offset)
        string_len = read_int(2)
        if is_utf8:
            string = f.read(string_len).decode("utf-8")
        else:
            string = f.read(string_len * 2).decode("utf-16-le")
        yield string


def main_apphash(args):
    env = UnityPy.Environment()
    app_package = None
    app_version = "unknown"
    app_hash = "unknown"

    if not args.ab_src:
        if not args.apk_src or args.fetch:
            from requests import get

            src = BytesIO()
            logger.debug("Fetching latest game package (JP) from APKPure")
            resp = get(
                "https://d.apkpure.net/b/XAPK/com.sega.pjsekai?version=latest",
                stream=True,
            )
            size = resp.headers.get("Content-Length", -1)
            with tqdm(total=int(size), unit="B", unit_scale=True) as progress:
                progress.sp = print
                for chunck in resp.iter_content(chunk_size=2**20):
                    src.write(chunck)
                    progress.update(len(chunck))
            src.seek(0)
        else:
            src = open(args.apk_src, "rb")
        with zipfile.ZipFile(src, "r") as zip_ref:
            manifests = [
                manifest
                for package in enum_package(zip_ref)
                for manifest in enum_candidates(
                    package, lambda fn: fn == "AndroidManifest.xml"
                )
            ]
            manifest = manifests[0][1]
            manifest_strings = set(dump_axml_stringpool(manifest))
            # Heur: Reverse lookup the package name from the manifest strings
            for ky in REGION_MAP:
                if ky in manifest_strings:
                    app_package = ky
                    break
            candidates = [
                candidate
                for package in enum_package(zip_ref)
                for candidate in enum_candidates(
                    package,
                    lambda fn: fn.split("/")[-1]
                    in {
                        "6350e2ec327334c8a9b7f494f344a761",  # PJSK Android
                        "c726e51b6fe37463685916a1687158dd",  # PJSK iOS
                        "data.unity3d",  # TW,KR,CN (ByteDance)
                    },
                )
            ]
            for candidate, stream, _ in candidates:
                env.load_file(stream)
    else:
        logger.info("Loading from AssetBundle %s" % args.ab_src)
        with open(args.ab_src, "rb") as f:
            env = load_assetbundle(BytesIO(f.read()))

    from sssekai.generated import TYPETREE_DEFS
    from sssekai.generated.Sekai import PlayerSettingConfig

    ans = None
    for pobj in env.objects:
        if pobj.type == UnityPy.enums.ClassIDType.MonoBehaviour:
            pname = pobj.peek_name()
            # Can't use UTTCGen_Reread because we may have no access to the Script PPtr asset
            # Object name seems to be a good enough heuristic
            for name in {"production_android", "production_ios"}:
                if pname == name:
                    # Works with post 3.4 (JP 3rd Anniversary) builds and downstream regional builds
                    tt = pobj.read_typetree(
                        TYPETREE_DEFS["Sekai.PlayerSettingConfig"], check_read=False
                    )
                    config = PlayerSettingConfig(**tt)
                    app_version = "%s.%s.%s" % (
                        config.clientMajorVersion,
                        config.clientMinorVersion,
                        config.clientBuildVersion,
                    )
                    data_version = "%s.%s.%s" % (
                        config.clientDataMajorVersion,
                        config.clientDataMinorVersion,
                        config.clientDataBuildVersion,
                    )
                    ab_version = "%s.%s.%s" % (
                        config.clientMajorVersion,
                        config.clientMinorVersion,
                        config.clientDataRevision,
                    )
                    app_hash = config.clientAppHash
                    # XXX: CN build has the same package name as the JP build??
                    # We need another heuristic to determine the package name
                    # This is previously done by setting app_package
                    package_heur = app_package or config.bundleIdentifier
                    region = region = REGION_MAP.get(package_heur, "unknown")
                    print(
                        f"Found {config.productName} at {config.m_Name}",
                        f"  Memo: {config.memo}",
                        f"  Package: {config.bundleIdentifier} (actually assumed as {package_heur})",
                        f"  AppHash (app_hash):     {config.clientAppHash}",
                        f"  Region  (app_region):   {region} (determined by {package_heur})",
                        f"  Version (app_version):  {app_version}",
                        f"  Version (ab_version):   {ab_version}",
                        f"  Bundle Version: {config.bundleVersion}",
                        f"  Data Version:   {data_version}",
                        f"  Version Suffix: {config.clientVersionSuffix}" "",
                        sep="\n",
                        file=sys.stderr,
                    )
                    # Only keep the Android one
                    if name == "production_android":
                        ans = f"""
# Auto-Generated by sssekai/entrypoint/apphash.py
# CLI Usage:
# sssekai abcache --app-platform android --app-region {region} --app-version {app_version} --app-appHash {app_hash} --app-abVersion {ab_version}
# Python Usage:
from sssekai.abcache import AbCacheConfig

AbCacheConfig(
    app_region="{region}",
    app_version="{app_version}",
    ab_version="{ab_version}",
    app_hash="{app_hash}",
    app_platform="android"
)"""
    print(ans)
